/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "threadSearch.h"
#include "search.h"
#include "moveGenerator.h"
#include "moveDo.h"
#include "hash.h"

#define MESSAGE_KILL 0
#define MESSAGE_START 1

#define HP 0
#define PVS 1
#define RPVS 2

struct tMessage{
	guint message;
	Position position;
	gint alpha;
	gint beta;
	gint depth;
	guint selDepth;
	gboolean nullMove;
	Move move;
	gint value;
	gint type;
};
typedef struct tMessage Message;

gint searchThreads = 1;
static GThread** wThread = NULL;
static GAsyncQueue* masterSlave;
static GAsyncQueue* slaveMaster;

gpointer wThreadStart(gpointer threadID);
void parallelSearch(Position* position, Move move, gint alpha, gint beta, gint depth, guint selDepth, 
		gboolean nullMove, gint type);
gint parallelSearchResult(Move* move, gint* alpha, gint* type);
void parallelSearchStop(gint idleThreads);

gint threadSearch(Position* position, gint alpha, gint beta, gint depth, guint selDepth, gboolean nullMove){
	Moves moves;
	Move move, bestMove = Empty, hMove = Empty, pMove=Empty;
	gboolean extend=FALSE, endGamePhase, fprune = FALSE;
	gint value, tmpDepth, fmargin = 0, movesSearched = 0, idleThreads, pAlpha, pType;
	guint64 key = position->key, hData;
	HashData* hd;

	if(stopFlag == TRUE && !ponderSearch) return EVAL_DRAW;

	searchNodes++;

	/* Draw */
	if(positionRepeat(position) >= 1 || position->staticMoves >= 100) return EVAL_DRAW;

	/* Hash */
	if(hashFind(&hd,key,&hData)){
		hMove = hashMove(hData);
        if(hashDepth(hData) >= depth){
			if(hashFlag(hData) == EXACT) return hashEval(hData);
			else if(hashFlag(hData) == ALPHA && hashEval(hData) > alpha){
				bestMove = hashMove(hData);
				alpha = hashEval(hData);
                if(alpha >= beta) return alpha;
			}
			else if(hashFlag(hData) == BETA && hashEval(hData) < beta){
				beta = hashEval(hData);
				if(alpha >= beta) return beta;
			}
		}
	}

	/* Extension */	
	if(position->extend){
		depth ++;
		extend = TRUE;
	}

	/* Horizon */
	if(depth < 1){
		value = quiescent(position,alpha,beta,selDepth+1);
		return value;
	}

	endGamePhase = positionEndGame(position);

	/* Null move pruning */
	if (nullMove && depth>2 && !extend && !endGamePhase){
		if(depth > 6) tmpDepth = depth - 4;
		else tmpDepth = depth - 3;
		moveDo(position,Null_Move);
		value = -threadSearch(position,-beta,-beta+1,tmpDepth,selDepth+1,FALSE);
		moveUndo(position);
		if(value >= beta){
			hashWrite(hd,key,depth,value,ALPHA,Empty);
			return value;
		}
	}

	/* Internal Iterative Deepening */
	tmpDepth = depth/2;
	if(hMove == Empty && depth > 3 && tmpDepth >= hashDepth(hData)){
		search(position,alpha, beta, tmpDepth, selDepth,FALSE);
		if(hashFind(&hd,key,&hData)) hMove = hashMove(hData);
	}

	/* Futility pruning */
	if(depth <= 3 && !extend && !endGamePhase){
		value = positionMaterial(position);
		switch(depth){
			case 3:{
				if(value + futilityValue3 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin3;
					depth --;
					break;
				}
			}
			case 2:{
				if(value + futilityValue2 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin2;
					break;
				}
			}
			case 1:{
				if(value + futilityValue1 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin1;
					break;
				}
			}
		}
	}

	/* First move */
	moves.state = START;
	if(hMove != Empty){
		move = hMove;
		moveDo(position,move);
	}
	else move = nextMoveDo(position,&moves);
	
	if(move != Empty){
		if(!(fprune && !position->extend && fmargin - positionMaterial(position) <= alpha)){
			value = -threadSearch(position, -beta, -alpha, depth-1, selDepth+1,TRUE);
			if(value > alpha){
				if(value >= beta){
					moveUndo(position);
					if(moves.state == STATIC) history[moveGetFrom(move)][moveGetTo(move)] += 1<<depth;
					hashWrite(hd,key,depth,value,ALPHA,move);
					return value;
				}
				alpha = value;
				bestMove = move;
			}
		}
		moveUndo(position);
		movesSearched ++;
	}

	/* Parallel search*/
	idleThreads = searchThreads;

	while(TRUE){
		if(idleThreads > 0){
			move = nextMoveDo(position,&moves);
			/* Last moves searched sequential */
			if(move == Empty){
				parallelStopFlag = TRUE;
				moves.size = moves.next = 0;
				while(idleThreads != searchThreads){
					parallelSearchResult(&pMove,&pAlpha,&pType);
					idleThreads++;
					moves.move[moves.size] = pMove;
					moves.size ++;
				}
				parallelStopFlag = FALSE;
				moves.toSort = moves.size;
				break;
			}
			/* Futility pruning */
			if(fprune && !position->extend && fmargin - positionMaterial(position) <= alpha){
				moveUndo(position);
				movesSearched ++;
				continue;
			}
			/* History pruning */
			if(movesSearched >= HistoryFullDepthMoves && depth > HistoryReductionLimit && 
				!extend && !position->extend){
				idleThreads--;
				parallelSearch(position, move, -(alpha+1), -alpha, depth-2, selDepth+1,TRUE,HP);
				moveUndo(position);
				movesSearched ++;
			}
			else{ /* PVS */
				idleThreads--;
				parallelSearch(position, move, -(alpha+1), -alpha, depth-1, selDepth+1,TRUE,PVS);
				moveUndo(position);
				movesSearched ++;
			}
		}
		else{
			value = -parallelSearchResult(&pMove,&pAlpha,&pType);
			idleThreads++;
			pAlpha = -pAlpha - 1;
			if(pType == HP && value > pAlpha){ /* FP */
				idleThreads--;
				moveDo(position,pMove);
				parallelSearch(position, pMove, -(pAlpha+1), -pAlpha, depth-1, selDepth+1,TRUE,PVS);
				moveUndo(position);
				continue;
			}
			if(pType == PVS && value > pAlpha && value < beta){ /* PVS research */
				idleThreads--;
				moveDo(position,pMove);
				parallelSearch(position, pMove, -beta, -pAlpha, depth-1, selDepth+1,TRUE,RPVS);
				moveUndo(position);
				continue;
			}
			if(value > alpha){
				if(value >= beta){
					parallelSearchStop(idleThreads);
					if(moveGetType(pMove) == Normal) history[moveGetFrom(pMove)][moveGetTo(pMove)] += 1<<depth;
					hashWrite(hd,key,depth,value,ALPHA,pMove);
					return value;
				}
				alpha = value;
				bestMove = pMove;
			}
		}
	}

	/* Last moves searched sequential */
	move = nextMoveDo(position,&moves);
	while(move != Empty){
		/* History pruning */
		if(movesSearched >= HistoryFullDepthMoves && depth > HistoryReductionLimit && 
			!extend && !position->extend){
			value = -threadSearch(position, -(alpha+1), -alpha, depth-2, selDepth+1,TRUE);
		}
		else value = alpha + 1;

		/* PVS */
		if(value > alpha){
			value = -threadSearch(position, -(alpha+1), -alpha, depth-1, selDepth+1,TRUE);
			if(value > alpha && value < beta)
				value = -threadSearch(position, -beta, -alpha, depth-1, selDepth+1,TRUE);
		}
		moveUndo(position);
		if(value > alpha){
			if(value >= beta){
				if(moveGetType(pMove) == Normal) history[moveGetFrom(move)][moveGetTo(move)] += 1<<depth;
				hashWrite(hd,key,depth,value,ALPHA,move);
				return value;
			}
			alpha = value;
			bestMove = move;
		}
		movesSearched ++;
		move = nextMoveDo(position,&moves);
	}

	/* Mate or pat */
	if(movesSearched == 0){
		if(position->inCheck == TRUE){
			value = -EVAL_MATE+selDepth;
			hashWrite(hd,key,depth,value,EXACT,Empty);
			return value;
		}
		else{  
			hashWrite(hd,key,depth,EVAL_DRAW,EXACT,Empty);
			return EVAL_DRAW;
		}
	}

	/* hash is update */
	if(bestMove != Empty) hashWrite(hd,key,depth,alpha,EXACT,bestMove);
	else hashWrite(hd,key,depth,alpha,BETA,hMove);

	return alpha;
}

void threadsCreate(){
	guint i;
	if(searchThreads == 1) return;
	masterSlave = g_async_queue_new();
	slaveMaster = g_async_queue_new();
	wThread = g_malloc(sizeof(GThread*)*searchThreads);
	for(i=0; i<searchThreads; i++)
		wThread[i] = g_thread_create(wThreadStart,NULL,TRUE,NULL);
}

void threadsDestroy(){
	guint i;
	stopFlag = TRUE;
	if(searchThreads == 1) return;
	Message* m = g_malloc(sizeof(Message));
	m->message = MESSAGE_KILL;
	for(i=0; i<searchThreads; i++) g_async_queue_push(masterSlave, (gpointer)m);
	for(i=0; i<searchThreads; i++) g_thread_join(wThread[i]);
	g_free(m);
	g_free(wThread);
	g_free(masterSlave);
	g_free(slaveMaster);
}

gpointer wThreadStart(gpointer threadID){
	Message* m;
	while(TRUE){
		m = (Message*)g_async_queue_pop(masterSlave);
		switch(m->message){
			case MESSAGE_KILL: { return NULL; }
			case MESSAGE_START: { 
				m->value = search(&m->position,m->alpha,m->beta,m->depth,m->selDepth,m->nullMove); 
				g_async_queue_push(slaveMaster, (gpointer)m);
				break;
			}
			default: {g_error("Thread message is wrong! (%u)\n",m->message); }

		}
	}
	return NULL;
}

void parallelSearch(Position* position, Move move, gint alpha, gint beta, gint depth, guint selDepth, 
	gboolean nullMove, gint type){
	Message* m = g_malloc(sizeof(Message));
	positionCopy(position,&m->position);
	m->move = move;
	m->alpha = alpha;
	m->beta = beta;
	m->depth = depth;
	m->selDepth = selDepth;
	m->nullMove = nullMove;
	m->message = MESSAGE_START;
	m->type = type;
	g_async_queue_push(masterSlave, (gpointer)m);
}

gint parallelSearchResult(Move* move,gint* alpha, gint* type){
	gint value;
	Message* m = (Message*)g_async_queue_pop(slaveMaster);
	value = m->value;
	*move = m->move;
	*alpha = m->alpha;
	*type = m->type;
	g_free(m);
	return value;
}

void parallelSearchStop(gint idleThreads){
	parallelStopFlag = TRUE;
	Move pMove;
	gint pAlpha, pType;
	while(idleThreads != searchThreads){
		parallelSearchResult(&pMove,&pAlpha,&pType);
		idleThreads++;
	}
	parallelStopFlag = FALSE;
}
