/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "uci.h"
#include "init.h"
#include "position.h"
#include "moveGenerator.h"
#include "moveDo.h"
#include "startSearch.h"
#include "search.h"
#include "threadSearch.h"
#include "hash.h"
#include "book.h"
#include "egbb.h"

#define SPIN 0
#define STRING 1
#define BUTTON 2
#define CHECK 3

static GMainLoop* loop;
static GIOChannel* uci_stdin;
static GRand* mrand;
static GTimer* timer;
static Position* position;
static SearchParameter searchParameter;
static guint DepthInfo = 5;
static guint MultiPVInfo = 5;
static guint CurrentMoveInfo = 10;
static GList* uciOption = NULL;
static gchar uciAbout[] = "http://labraj.uni-mb.si/en/index.php/Chess_program_BBChess";
static gboolean ponder;

struct tUciOption{
	gint def;
	gchar sDef[255];
	gint min;
	gint max;
	gchar name[255];
	gpointer value;
	gint type;
};
typedef struct tUciOption UCIOption;

gboolean uciReadLine(GIOChannel* uci_stdin, GIOCondition condition, gpointer data);
void stringMovesDo(Position* position, gchar moves[]);
void uciAddSpinOption(const gchar name[255], const gint min, const gint max, gpointer value);
void uciAddStringOption(const gchar name[255], const gchar def[255], gpointer value);
void uciAddButtonOption(const gchar name[255]);
void uciAddCheckOption(const gchar name[255], gboolean def, gpointer value);
void uciIdentity(void);
void uciIsReady(void);
void uciQuit(void);
void uciGo(gchar line[]);
void uciStop(void);
void uciSetPosition(gchar line[]);
void uciSetOption(gchar line[]);
void uciNewGame(void);
void uciDebug(void);
void uciOptionsInit(void);

void uciRun(void){
	mrand = g_rand_new_with_seed(0);
	timer = g_timer_new();
	loop = g_main_loop_new(NULL,FALSE);
	init(mrand);	
	position = positionNew();
	uci_stdin = g_io_channel_unix_new(0);
	if(!uci_stdin) g_error("Cannot create new uci GIOChannel!\n");
	if(!g_io_add_watch(uci_stdin,G_IO_IN, uciReadLine, NULL))
		g_error("Cannot add watch on GIOChannel!\n");
	if(!g_thread_supported()) g_thread_init(0);
	else g_error("Threads not suported !!!\n");	
	uciOptionsInit();
	hashCreate();
	pawnHashCreate();  	 	 
	threadsCreate();
	egbbLoadLibrary();
	bookOpen();
	g_main_loop_run(loop);
}

gboolean uciReadLine(GIOChannel* uci_stdin, GIOCondition condition, gpointer data){
	gchar command[20];
	gchar* subline, *line;
	g_io_channel_read_line(uci_stdin, &line, NULL, NULL,NULL);
	sscanf(line,"%s",command);
	subline = g_strstr_len(line,strlen(line), command) + strlen(command);
	if(!g_ascii_strcasecmp(command, "stop")) uciStop();
	else if(!g_ascii_strcasecmp(command, "ponderhit")) ponderSearch = FALSE;
	else if(!g_ascii_strcasecmp(command, "go")) uciGo(subline);
	else if(!g_ascii_strcasecmp(command, "position")) uciSetPosition(subline);
	else if(!g_ascii_strcasecmp(command, "setoption")) uciSetOption(subline);
	else if(!g_ascii_strcasecmp(command, "ucinewgame"))	uciNewGame();
	else if(!g_ascii_strcasecmp(command, "isready")) uciIsReady();
	else if(!g_ascii_strcasecmp(command, "quit")) uciQuit();
	else if(!g_ascii_strcasecmp(command, "uci")) uciIdentity();
	else if(!g_ascii_strcasecmp(command, "debug")) uciDebug();
	else { g_printf("Unknown uci command:%s\n",command); fflush(stdout); }
	g_free(line);
	return TRUE;
}

void uciIdentity(void){
	guint i;
	UCIOption* option;
	g_printf("id name %s\n",ENGINE);
	g_printf("id author Borko Boskovic\n");
	for(i=0; i<g_list_length(uciOption); i++){
		option = g_list_nth_data(uciOption,i);
		switch(option->type){
			case SPIN:{
				g_printf("option name %s type spin default %i min %i max %i\n",
					option->name, option->def, option->min, option->max);
				break;
			}
			case STRING:{
				g_printf("option name %s type string default %s\n", option->name, option->sDef);
				break;
			}
			case BUTTON:{
				g_printf("option name %s type button\n", option->name);
				break;
			}
			case CHECK:{
				if(option->def)	g_printf("option name %s type check default true\n", option->name);
				else g_printf("option name %s type check default false\n", option->name);
				break;
			}
		}
	}
	g_printf("uciok\n");
	fflush(stdout);
}

void uciQuit(void){
	g_io_channel_shutdown(uci_stdin,TRUE,NULL);
	g_main_loop_quit(loop);
	g_main_loop_unref(loop);
	positionFree(position);
	g_rand_free(mrand);
	g_timer_destroy(timer);
	g_list_free(uciOption);
	bookClose();
	egbbClose();
	threadsDestroy();
	hashDestroy();
	pawnHashDestroy(); 		 
}

void uciGo(gchar line[]){
	gchar *subline;
	searchParameter.depth = searchParameter.movetime = searchParameter.movestogo = 0;
   	searchParameter.nodes = 0;
	searchParameter.wtime = searchParameter.btime = 0;
	searchParameter.winc = searchParameter.binc = 0;
	searchParameter.infinite = searchParameter.mate = FALSE;

	subline = g_strstr_len(line, strlen(line), "depth");
	if(subline != NULL) sscanf(subline,"depth %u",&searchParameter.depth);
	subline = g_strstr_len(line, strlen(line), "movetime");
	if(subline != NULL) sscanf(subline,"movetime %u",&searchParameter.movetime);
	subline = g_strstr_len(line, strlen(line), "nodes");
	if(subline != NULL) sscanf(subline,"nodes %llu",&searchParameter.nodes);
	subline = g_strstr_len(line, strlen(line), "infinite");
	if(subline != NULL) searchParameter.infinite = TRUE;
	subline = g_strstr_len(line, strlen(line), "mate");
	if(subline != NULL) searchParameter.mate = TRUE;
	subline = g_strstr_len(line, strlen(line), "wtime");
	if(subline != NULL) sscanf(subline,"wtime %u",&searchParameter.wtime);
	subline = g_strstr_len(line, strlen(line), "btime");
	if(subline != NULL) sscanf(subline,"btime %u",&searchParameter.btime);
	subline = g_strstr_len(line, strlen(line), "winc");
	if(subline != NULL) sscanf(subline,"winc %u",&searchParameter.winc);
	subline = g_strstr_len(line, strlen(line), "binc");
	if(subline != NULL) sscanf(subline,"binc %u",&searchParameter.binc);
	subline = g_strstr_len(line, strlen(line), "movestogo");
	if(subline != NULL) sscanf(subline,"movestogo %u",&searchParameter.movestogo);
	subline = g_strstr_len(line, strlen(line), "ponder");
	if(subline != NULL) ponderSearch = TRUE;
	else ponderSearch = FALSE;

	g_timer_start(timer);
	searchStop();
	searchStart(position,&searchParameter);
}

void uciStop(void){
	ponderSearch = FALSE;
	searchStop();
}

void uciSetPosition(gchar line[]){
	gchar command[20];
	sscanf(line,"%s",command);
	if(!g_ascii_strcasecmp(command, "startpos")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		sscanf(line,"%s",command);
		positionStartPosition(position);
		if(!g_ascii_strcasecmp(command, "moves")){
			line = g_strstr_len(line, strlen(line), command) + strlen(command);
			stringMovesDo(position,line);
		}
	}
	else if(!g_ascii_strcasecmp(command, "fen")){
		line = g_strstr_len(line, strlen(line), command) + strlen(command);
		positionSetFen(position,line);
		line = g_strstr_len(line, strlen(line), "moves");
		if(line != NULL){
			line = line + 6;
			stringMovesDo(position,line);
		}
	}
}

void uciOptionsInit(void){
	uciAddCheckOption("Ponder", ponder, &ponder);
	uciAddSpinOption("MultiPV" , 1, 20, &searchMultipv);
	uciAddSpinOption("Threads", 1, 16, &searchThreads);
	uciAddSpinOption("Hash", 1, 4096, &hashSize);
	uciAddCheckOption("OwnBook", ownBook, &ownBook);
	uciAddStringOption("Book File", bookPath, bookPath);
    uciAddStringOption("Bitbases Path", bitbasesPath, bitbasesPath);
	uciAddSpinOption("Bitbases Cache Size", 1,256, &egbbCacheSize);
	uciAddButtonOption("Clear Hash");
    uciAddStringOption("UCI_EngineAbout", uciAbout, uciAbout);
}

void uciSetOption(gchar line[]){
	gchar tmp[255], name[255]="";
	gint i,value;
	gboolean found = FALSE;
	UCIOption* option = NULL;
	sscanf(line," name %s ",name);
	line = g_strstr_len(line, strlen(line), name) + strlen(name);
	sscanf(line," %s ",tmp);
	while(g_ascii_strcasecmp(tmp,"value")){
		strcat(name," ");
		strcat(name,tmp);
		line = g_strstr_len(line, strlen(line), tmp) + strlen(tmp);
		if(EOF == sscanf(line," %s ",tmp)) break;
	}
	for(i=0; i< g_list_length(uciOption); i++){
		option = g_list_nth_data(uciOption,i);
		if(!g_ascii_strcasecmp(name,option->name)){
			switch(option->type){
				case SPIN:{
					sscanf(line," value %d",&value);
					if(!g_ascii_strcasecmp(name,"Threads"))
						threadsDestroy();
					if(value >= option->min && value <= option->max){
						   	*(gint*)option->value = value;
						   	g_printf("info string %s set to %d\n",option->name,value);
							fflush(stdout);
					}
					else g_error("Wrong uci option value (%s:%d)\n",option->name,value);
					if(!g_ascii_strcasecmp(name,"Threads"))
						threadsCreate();
					break;
				}
				case STRING:{
					sscanf(line," value %s",(gchar*)option->value);
					break;
				}
				case BUTTON:{
					break;
				}
				case CHECK:{
					sscanf(line," value %s",tmp);
					if(!g_ascii_strcasecmp(tmp,"true")) *(gint*)option->value = TRUE;
					else if(!g_ascii_strcasecmp(tmp,"false")) *(gint*)option->value = FALSE;
					else g_error("Wrong value %s!\n",tmp);
					g_printf("info string %s set to %s\n",option->name,tmp);
					fflush(stdout);
					break;
				}
			}
			found = TRUE;
			break;
		}
	}
	if(!found) g_printf("info wrong uci option name (%s)\n",name);

	if(!g_ascii_strcasecmp(name,"Clear Hash")){
		hashReset();
	}
	else if(!g_ascii_strcasecmp(name,"Hash")){
		hashDestroy();
		hashCreate();
	}
	else if(!g_ascii_strcasecmp(name,"Book File")){
		bookClose();
		bookOpen();
	}
	else if(!g_ascii_strcasecmp(name,"Bitbases Path") || !g_ascii_strcasecmp(name,"Bitbases Cache Size")){
		egbbLoadLibrary();
	}
}

void uciNewGame(void){
}

void uciIsReady(void){
	g_printf("readyok\n");
	fflush(stdout);
}


void uciDebug(void){
	positionPrint(position);
	fflush(stdout);
}

void uciCurrentMove(const guint depth, const guint index, const RootMoves* moves){
	if(depth < CurrentMoveInfo) return;
	gchar sMove[6];
	moveToString(moves->move[index].move,sMove);
	g_printf("info currmovenumber %d currmove %s\n",index+1,sMove);
	fflush(stdout);
}

void uciMultiPv(const Position* position, const guint depth, const guint searchMultipv, const RootMoves* moves){
	guint i,j;
	if(depth < MultiPVInfo) return;
	for(i=0; i<searchMultipv; i++){
		g_printf("info ");
		if(searchMultipv > 1) g_printf("multipv %d ",i+1);
		if(moves->move[i].eval > EVAL_MATE - 40){
			if(position->turn != White)
				g_printf("score mate %d ",(EVAL_MATE - moves->move[i].eval+2)/2);
			else
				g_printf("score mate %d ",(EVAL_MATE - moves->move[i].eval+3)/2);
		}
		else if(moves->move[i].eval < -EVAL_MATE + 40){
			if(position->turn != White)
				g_printf("score mate %d ",(-EVAL_MATE - moves->move[i].eval -1)/2);
			else
				g_printf("score mate %d ",(-EVAL_MATE - moves->move[i].eval -2)/2);
		}
		else g_printf("score cp %d ",moves->move[i].eval);
		if(bbhits != 0) g_printf("bbhits %d ",bbhits);
		g_printf("depth %d seldepth %d time %d nodes %llu nps %u hashfull %d pv ",
			moves->move[i].depth, moves->move[i].seldepth, moves->move[i].time, 
			(unsigned long long)moves->move[i].nodes, moves->move[i].nps, hashUsage());
		movePrint(moves->move[i].move);
		j=0;
		while(moves->move[i].pv[j] != Empty) {
			movePrint(moves->move[i].pv[j]);
			j++;
		}
		g_printf("\n");
	}
	fflush(stdout);
}

void uciBeginDepthInfo(guint depth){
	if(depth >= DepthInfo){
		g_printf("info depth %d\n",depth);
		fflush(stdout);
	}
}

void uciEndDepthInfo(const guint depth, const guint64 searchNodes){
	gdouble sec;
	guint nps;
	if(depth >= DepthInfo){
		sec = g_timer_elapsed(timer,NULL);
		if(sec > 0) nps = (guint)(searchNodes/(gfloat)sec);
		else nps = 0;
		g_printf("info time %u nodes %llu nps %u hashfull %u \n", 
					(guint)(sec*1000), (unsigned long long)searchNodes,nps,hashUsage());
		fflush(stdout);
	}
}

void uciBestMove(const Move move, const Move pMove){
	g_printf("bestmove ");
	movePrint(move);
	if(ponder && pMove != Empty){
		g_printf("ponder ");
		movePrint(pMove);
	}
	g_printf("\n");
	fflush(stdout);
}

void uciAddSpinOption(const gchar name[255], const gint min, const gint max, gpointer value){
	UCIOption* no = g_malloc(sizeof(UCIOption));
	strcpy(no->name,name);
	no->def = *(gint *)value;
	no->min = min;
	no->max = max;
	no->value = value;
	no->type = SPIN;
	uciOption = g_list_append(uciOption,no);
}

void uciAddStringOption(const gchar name[255], const gchar def[255], gpointer value){
	UCIOption* no = g_malloc(sizeof(UCIOption));
	strcpy(no->name,name);
	strcpy(no->sDef,def);
	no->value = value;
	no->type = STRING;
	uciOption = g_list_append(uciOption,no);
}

void uciAddButtonOption(const gchar name[255]){
	UCIOption* no = g_malloc(sizeof(UCIOption));
	strcpy(no->name,name);
	no->type = BUTTON;
	uciOption = g_list_append(uciOption,no);
}

void uciAddCheckOption(const gchar name[255], gboolean def, gpointer value){
	UCIOption* no = g_malloc(sizeof(UCIOption));
	strcpy(no->name,name);
	no->def = (gint)def;
	no->type = CHECK;
	no->value = value;
	uciOption = g_list_append(uciOption,no);
}

double uciGetTime(void){
	return g_timer_elapsed(timer,NULL);
}
